import logging

from django.http import JsonResponse
from telethon.errors import RPCError, SessionPasswordNeededError

from .exceptions import PayloadException, TelegramAuthorizationException
from .helpers import Telegram, require_post, login_required, parse_json_payload
from .models import TelegramAuthorization

logger = logging.getLogger("django-telethon-authorization")


@login_required
@require_post
def request_code(request):
    """ Accept JSON {phone: +xxxxxxxxxxx} """
    try:
        phone, = parse_json_payload(request.body, "phone")
    except PayloadException as e:
        logger.warning(e)
        return e.to_response()

    auth, _ = TelegramAuthorization.objects.get_or_create(user=request.user)

    try:
        client = Telegram.get_client(phone)
    except TelegramAuthorizationException as e:
        return e.to_response()

    if client.is_user_authorized():
        return JsonResponse({"success": False, "message": "You are already authorized"})
    try:
        response = client.send_code_request(phone)
        # hash will be needed during code submission
        auth.phone_code_hash = response.phone_code_hash
        # if user login with another account - change phone of TelegramAuthorization object
        auth.phone = phone
        auth.save()
        client.disconnect()
        return JsonResponse({"success": True})

    except RPCError as e:
        return JsonResponse(
            {"success": False, "message": "Telegram exception occurred. %s. %s. %s" % (e.code, e.message, str(e))})

    except Exception as e:
        logger.exception("TG REQUEST CODE. POST. Error occurred during telegram send code")
        return JsonResponse({"success": False, "message": "'Error occurred during code sending\n%s'" % e})


@login_required
@require_post
def submit(request):
    try:
        phone, code, password = parse_json_payload(request.body, "phone", "code", "password")
    except PayloadException as e:
        logger.exception(e)
        return e.to_response()

    try:
        auth = TelegramAuthorization.objects.get(user=request.user, phone=phone)
    except TelegramAuthorization.DoesNotExist:
        return JsonResponse({"success": False, "message": "Phone '%s' is invalid'" % phone})

    client = Telegram.get_client(phone)

    try:
        client.sign_in(auth.phone, code, phone_code_hash=auth.phone_code_hash)
    except SessionPasswordNeededError:
        if password:
            client.sign_in(password=password)
        else:
            return JsonResponse({
                "success": False, "message": "Two Factor Authorization enabled. Please provide both code and password"
            })

    except RPCError as e:
        return JsonResponse(
            {"success": False, "message": "Telegram exception occurred. %s. %s. %s" % (e.code, e.message, str(e))})

    except Exception as e:
        logger.warning("TG Login. POST. Error occurred during telegram sign-in\n%s" % e)
        return JsonResponse({"success": False, "message": "'Error occurred during telegram sign-in\n%s'" % e})

    client.disconnect()

    # do not store hash after successful login
    auth.phone_code_hash = None
    auth.save()

    return JsonResponse({"success": True})


@login_required
@require_post
def logout(request):
    try:
        phone, = parse_json_payload(request.body, "phone")
    except PayloadException as e:
        logger.warning(e)
        return e.to_response()

    try:
        telegram_authorization = TelegramAuthorization.objects.get(user=request.user, phone=phone)
    except TelegramAuthorization.DoesNotExist:
        return JsonResponse({"success": False, "message": "Phone '%s' is invalid'" % phone})

    client = Telegram.get_client(telegram_authorization.phone)

    if client.log_out():
        # delete auth record
        telegram_authorization.delete()
        return JsonResponse({"success": True})
    else:
        return JsonResponse({"success": False, "message": "Telegram RPC error"})



# Generated by Django 2.0.6 on 2018-08-04 14:31

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('django_telethon_authorization', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='telegramauthorization',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterUniqueTogether(
            name='telegramauthorization',
            unique_together={('user', 'phone')},
        ),
    ]



# Generated by Django 2.0.6 on 2018-08-04 14:50

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_telethon_authorization', '0002_auto_20180804_1431'),
    ]

    operations = [
        migrations.AlterField(
            model_name='telegramauthorization',
            name='user',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterUniqueTogether(
            name='telegramauthorization',
            unique_together=set(),
        ),
    ]



import json
from functools import wraps
from typing import Union

import os

from django.conf import settings
from django.http import JsonResponse, HttpRequest
from telethon import TelegramClient
from telethon.sessions import SQLiteSession

from .models import TelegramAuthorization
from .exceptions import TelegramAuthorizationException, PayloadException


class Telegram:

    @staticmethod
    def get_client(request_or_phone: Union[HttpRequest, str], session_class=SQLiteSession) -> TelegramClient:
        """
        Create and return prepared telegram client based request or phone
        """
        session_path = settings.TG_SESSION_PATH
        api_id = settings.TG_API_ID
        api_hash = settings.TG_API_HASH

        if isinstance(request_or_phone, HttpRequest):
            user = request_or_phone.user
            try:
                telegram_authorization = TelegramAuthorization.objects.get(user=user)
            except TelegramAuthorization.DoesNotExist:
                raise TelegramAuthorizationException("User doesn't have valid telegram session")
            phone = telegram_authorization.phone
        else:
            phone = request_or_phone

        if session_class == SQLiteSession:
            # default behavior
            if not session_path:
                raise TelegramAuthorizationException(
                    "SQLite session implementation used, but variable TG_SESSION_PATH was not set."
                )
            session = SQLiteSession(os.path.join(session_path, "%s.session" % phone))
        else:
            session = session_class(phone)

        client = TelegramClient(
            session,
            api_id,
            api_hash,
            spawn_read_thread=False,
            report_errors=False
        )
        connected = client.connect()
        if not connected:
            raise TelegramAuthorizationException("Can't connect to telegram servers")

        return client

    @staticmethod
    def is_authorized(request) -> bool:
        """
        Determine whether user has authorized session based on HttpRequest
        """
        try:
            client = Telegram.get_client(request)
        except TelegramAuthorizationException:
            return False

        return client.is_user_authorized()


def parse_json_payload(body, *keys):
    """
    Parse request.body and yield lookup values
    """
    try:
        raw_payload = body.decode("utf-8")
    except UnicodeDecodeError as exc:
        raise PayloadException("Cant decode body '%s'\n%s" % (body, exc))
    try:
        payload = json.loads(raw_payload)
    except (ValueError, TypeError) as exc:
        raise PayloadException("Can't load JSON from raw payload '%s'\n%s" % (raw_payload, exc))
    for key in keys:
        yield payload.get(key)


def require_post(func):
    @wraps(func)
    def wrapper(request, *args, **kwargs):
        if request.method == "POST":
            return func(request, *args, **kwargs)
        return JsonResponse({"success": False, "message": "Only POST requests allowed"})

    return wrapper


def login_required(func):
    @wraps(func)
    def wrapper(request, *args, **kwargs):
        if request.user.is_authenticated:
            return func(request, *args, **kwargs)
        return JsonResponse({"success": False, "message": "Authentication required"})

    return wrapper
